# -*- coding:utf-8 -*-
# Author: hankcs
# Date: 2019-12-31 23:54
from hanlp_common.constant import HANLP_URL

SEMEVAL16_NEWS_BIAFFINE_ZH = HANLP_URL + 'sdp/semeval16-news-biaffine_20191231_235407.zip'
'Biaffine SDP (:cite:`he-choi-2019`) trained on SemEval16 news data.'
SEMEVAL16_TEXT_BIAFFINE_ZH = HANLP_URL + 'sdp/semeval16-text-biaffine_20200101_002257.zip'
'Biaffine SDP (:cite:`he-choi-2019`) trained on SemEval16 text data.'

SEMEVAL16_ALL_ELECTRA_SMALL_ZH = HANLP_URL + 'sdp/semeval16_sdp_electra_small_20220218_140116.zip'
'Biaffine SDP (:cite:`he-choi-2019`) trained on SemEval16 text and news data. UF=80.82% LF=70.02%.'

SEMEVAL15_PAS_BIAFFINE_EN = HANLP_URL + 'sdp/semeval15_biaffine_pas_20200103_152405.zip'
'Biaffine SDP (:cite:`he-choi-2019`) trained on SemEval15 PAS data.'
SEMEVAL15_PSD_BIAFFINE_EN = HANLP_URL + 'sdp/semeval15_biaffine_psd_20200106_123009.zip'
'Biaffine SDP (:cite:`he-choi-2019`) trained on SemEval15 PSD data.'
SEMEVAL15_DM_BIAFFINE_EN = HANLP_URL + 'sdp/semeval15_biaffine_dm_20200106_122808.zip'
'Biaffine SDP (:cite:`he-choi-2019`) trained on SemEval15 DM data.'

ALL = {}
